/*
 * Wacom USB digitizers
 */

#include <u.h>
#include <libc.h>
#include <thread.h>
#include "usb.h"
#include "wacom.h"

typedef struct WDev WDev;
typedef struct WOut WOut;

enum {
	Bleft = 1,
	Bmiddle = 2,
	Bright = 4,
	Bdown = 8,
	Bup = 16,
	Bright2 = 32,
	Bleft2 = 64,

	IDpen = 0x85,
	IDmouse4 = 0x09,
};

enum {
	OutXYZ = 1<<0,
	OutPress = 1<<3,
	OutTilt = 1<<4,
	OutRotate = 1<<5,
	OutButton = 1<<6,
	OutWheel = 1<<7,

	OutPen = OutXYZ|OutPress|OutTilt|OutButton,
	OutMouse4 = OutXYZ|OutButton|OutWheel,
	OutMouse4a = OutXYZ|OutRotate,
};

struct WDev
{
	Dev*	dev;		/* usb device*/
	Dev*	ep;		/* endpoint to get events */
	int	accel;		/* only for mouse */
};

struct WOut
{
	char *tool;
	ulong	enable;
	uint	x, y, z;
	uint	buttons;
	uint	rotation;
	uint	pressure;
	int	xtilt, ytilt;
	int	wheel;
};

struct WMouse
{
	uint x, y, buttons;
	int enable;
};

enum {
	TBintous = 1,
};

Cinfo
wacominfo[] =
{
	{0x056a, 0x0041, TBintous},	/* Intous2 4x5 */
	{0x056a, 0x0042, TBintous},	/* Intous2 6x8 */
	{0x056a, 0x0043, TBintous},	/* Intous2 9x12 */
	{0x056a, 0x0044, TBintous},	/* Intous2 12x12 */
	{0x056a, 0x0045, TBintous},	/* Intous2 12x18 */
	{0, 0, 0},
};

static int wacomdebug;

static int mouseinfd;

static int
setfirstconfig(WDev* f, int eid, uchar *, int )
{
	int nr, id;

	dprint(2, "setting first config\n");

	id = f->dev->usb->ep[eid]->iface->id;

	nr =usbcmd(f->dev,  Rh2d | Rstd | Rdev, Rsetconf, 1, id, nil, 0);
	if(nr < 0)
		return -1;

	nr=usbcmd(f->dev, Rh2d | Rclass | Riface, Setidle,  0, id, nil, 0);
	if(nr < 0)
		return -1;

	nr=usbcmd(f->dev, Rh2d | Rclass | Riface, Setreport, 3<<8 | 2, id, (uchar *)"\2\2", 2);
	if(nr < 0)
		return -1;

	return nr;
}

static void
wacomfatal(WDev *wd, char *sts)
{
	Dev *dev;

	if(sts != nil)
		fprint(2, "wacom: fatal: %s\n", sts);
	else
		fprint(2, "wacom: exiting\n");
	dev = wd->dev;
	wd->dev = nil;
	if(wd->ep != nil)
		closedev(wd->ep);
	wd->ep = nil;
	devctl(dev, "detach");
	closedev(dev);
	/*
	 * free(wd); done by closedev.
	 */
	threadexits(sts);
}

static void
sethipri(void)
{
	char fn[30];
	int fd;

	snprint(fn, sizeof fn, "/proc/%d/ctl", getpid());
	fd = open(fn, OWRITE);
	if(fd >= 0) {
		fprint(fd, "pri 13");
		close(fd);
	}
}

static uchar
getbit(uchar *buf, int n)
{
	int nbi, nby, r;
	nbi = 7-(n % 8);
	nby = n / 8;
	r = buf[nby];
	return (r>>nbi) & 1;
}

static ulong
getbits(uchar *buf, int n, int nbits)
{
	int i;
	ulong r;

	r = 0;
	for(i = 0; i < nbits; i++){
		r <<= 1;
		r |= getbit(buf, n+i);
	}
	return r;
}

static void
domouse(struct WMouse *m, WOut *o){
	int dx, dy, db;
	char mbuf[50];

	if(m->enable){
		dx = ((int)o->x - (int)m->x)/8;
		dy = ((int)o->y - (int)m->y)/8;
		db = o->buttons ^ m->buttons;
		if(dx || dy || db || o->buttons & 0x18){
			seprint(mbuf, mbuf+sizeof(mbuf), "m%11d %11d %11d",
				dx, dy, o->buttons);
			dprint(2,"%s\n", mbuf);
			write(mouseinfd, mbuf, strlen(mbuf));
		}
	}
	m->enable = 1;
	m->x = o->x;
	m->y = o->y;
	m->buttons = o->buttons;
};

static void
wacomwork(void* a)
{
	int tool, inrange, extra, altp, eraser, c, ptrfd;
	uchar buf[MaxPktLen];
	WDev* f = a;
	WOut o, mouse4;
	struct WMouse mpen, mm4;

	if(f->ep == nil)
		wacomfatal(f, nil);
	ptrfd = f->ep->dfd;
	if(f->ep->maxpkt < 3 || f->ep->maxpkt > MaxPktLen)
		wacomfatal(f, "weird maxpkt");
	sethipri();
	eraser = 0;
	memset(&mouse4, 0, sizeof(WOut));
	for(;;){
		memset(buf, 0, MaxPktLen);
		c = read(ptrfd, buf, f->ep->maxpkt);
		assert(f->dev != nil);
		if(c <= 0){
			dprint(2, "wacom: %s: read: %r\n", f->ep->dir);
			wacomfatal(f, nil);
		}
		inrange = getbit(buf, 9);
		tool = getbits(buf, 10, 3);	/* 3 bits of tool, 3 bits of option */
		extra = getbits(buf, 13, 2);
		altp = getbit(buf, 15);
		o.x = getbits(buf, 16, 16);
		o.y = getbits(buf, 32, 16);
		o.z = getbits(buf, 72, 5);
		if(!inrange) o.z = 0xff;	/* fixme: find a better value */
		switch(tool){
		case 0:	/* clear */
			if(extra == 1){	/* identifier */
				switch(getbits(buf, 16, 8)){
				case IDpen:
					mpen.enable = 0;
					eraser = getbit(buf, 24);	/* yay magic */
					break;
				case IDmouse4:
					mm4.enable = 0;
					break;
				}
				dprint(1, "id	%.*H\n", c-2, buf + 2);
			}
			break;
		case 4:
			o.tool = "pen";
			o.enable = OutPen;
			o.pressure = getbits(buf, 48, 10);
			o.xtilt = getbits(buf, 58, 7) - 0x40;
			o.ytilt = getbits(buf, 65, 7) - 0x40;
			o.buttons = getbits(buf, 13, 2) << 1;
			if(o.pressure > 0x200) o.buttons |= 1;
			if(eraser) o.buttons |= Bright2;
			domouse(&mpen, &o);
			dprint(2, "pen	A %04x %04x %02x	P %03x	T %02x %02x	B %x\n",
				o.x, o.y, o.z, o.pressure, o.xtilt, o.ytilt, o.buttons);
			break;
		case 5:	/* 4d mouse */
			o.tool = "m4";
			switch(extra){
			case 0:
				o.enable = OutMouse4;
				o.wheel = getbits(buf, 48, 10);
				if(getbit(buf, 68)) o.wheel *= -1;
				o.buttons = getbits(buf, 66, 2) << 5;
				if(o.wheel > 0) o.buttons |= 1<<4;
				if(o.wheel < 0) o.buttons |= 1<<3;
				o.buttons |= getbits(buf, 69, 3);
				memmove(&mouse4, &o, sizeof(WOut));
				break;
			case 1:
				memmove(&o, &mouse4, sizeof(WOut));
				if(o.enable != OutMouse4) break;
				o.enable |= OutMouse4a;
				o.rotation = getbits(buf, 48, 10);
				if(!getbit(buf, 58)) o.rotation *= -1;
				domouse(&mm4, &o);
				dprint(2, "%s	A %04x %04x %02x	R %x	W %x	B %x\n",
					o.tool, o.x, o.y, o.z, o.rotation, o.wheel, o.buttons);
				break;
			}
			break;
		default:
			dprint(2, "tool %x %x %x	%04x %04x %02x %.*H\n",
				tool, extra, altp, o.x, o.y, o.z, c-1, buf + 1);
			break;
		}
	}
}

static void
freewdev(void *a)
{
	WDev *wd;

	wd = a;
	dprint(2, "freewdev\n");
	free(wd);
}

static void
wacomstart(Dev *d, Ep *ep, void (*f)(void*), WDev *wd)
{
	uchar desc[128];
	int res;

	d->free = freewdev;
	wd->dev = d;

	res = setfirstconfig(wd, ep->id, desc, sizeof desc);
	if(res < 0)
		return;

	wd->ep = openep(d, ep->id);
	if(wd->ep == nil){
		fprint(2, "wacom: %s: openep %d: %r\n", d->dir, ep->id);
		return;
	}
	if(opendevdata(wd->ep, OREAD) < 0){
		fprint(2, "wacom: %s: opendevdata: %r\n", wd->ep->dir);
		closedev(wd->ep);
		wd->ep = nil;
		return;
	}

	incref(d);
	proccreate(f, wd, Stack);
}

static int
usage(void)
{
	werrstr("usage: usb/wacom [-d] [-a n] [-N nb]");
	return -1;
}

int
wacommain(Dev *d, int argc, char* argv[])
{
	int i, accel, devid;
	Ep *ep;
	WDev *wd;
	Usbdev *ud;
	Cinfo *cp;

	fmtinstall('H', encodefmt);
	accel = 0;
	devid = d->id;
	ARGBEGIN{
	case 'a':
		accel = strtol(EARGF(usage()), nil, 0);
		break;
	case 'd':
		wacomdebug++;
		break;
	case 'N':
		devid = atoi(EARGF(usage()));		/* ignore dev number */
		break;
	default:
		return usage();
	}ARGEND;
	if(argc != 0)
		return usage();
	USED(devid);
	ud = d->usb;
	d->aux = nil;
	dprint(2, "wacom: main: dev %s ref %ld\n", d->dir, d->ref);
	mouseinfd = open("/dev/mousein", OWRITE);

	for(i = 0; i < nelem(ud->ep); i++){
		if((ep = ud->ep[i]) == nil)
			break;
		if(ep->type != Eintr)
			continue;
		for(cp=wacominfo; cp->vid != 0; cp++){
			if(ud->vid == cp->vid && ud->did == cp->did){
				wd = d->aux = emallocz(sizeof(WDev), 1);
				wd->accel = accel;
				wacomstart(d, ep, wacomwork, wd);
			}
		}
	}
	return 0;
}
